/**
 * @author: Ben Y. Zhao (ravenben@cs.berkeley.edu)
 * @version: Release 1.0 $date:  March 1, 1999$
 * (c) 1999 by Ben Yanbin Zhao and the Regents of the University of California
 * 
 * This is the code release of the XSet XML search engine on treaps.  The XSet
 * uses treaps to represent hierarchies of tags and indices on tag values.  
 * An ongoing progress homepage of the project can be found at
 * http://www.cs.berkeley.edu/~ravenben/ninja/xset.html
 * SETserver.java is the main Server code for XSet
 */

package xset;

import java.util.*;
import java.io.*;
import com.ms.xml.om.*;
import com.ms.xml.util.*;

/**
 * Class that provides the main interface to the Search Engine on Treaps
 * It provides the public methods register and deregister for the
 * external applications.
 */
public class SETserver {

  public static final short HASHSTORE = 0;
  public static final short TREAPSTORE = 1;

  protected Hashtable rootTagList;
  protected boolean refresh = false;
  protected long timeout = 0;
  protected Thread runCleaner = null;
  protected Hashtable globalHashSID = null;;

  /**
   * Constructor method
   * Initializes variables and prepares the search indices for data
   * @param refreshRate The time taken for an entry to become outdated
   */
  public SETserver(long refreshRate) {
    rootTagList = new Hashtable();
    globalHashSID = new Hashtable();
    if (refreshRate > 0) {
      timeout = refreshRate;
      refresh = true;
      runCleaner = new Thread(new SETcleaner(this, timeout));
      runCleaner.start();
    } else {
      System.err.println("ERROR: refreshRate < 0!");
      System.exit(-1);
    }
  }

  /**
   * Constructor method
   * Initializes variables and prepares the search indices for data
   * This constructor is for SET instances with no timeouts
   */
  public SETserver() {
    rootTagList = new Hashtable();
    globalHashSID = new Hashtable();
  }

  /**
   * Method called by the SDS server or independent application to
   * query for a certain document, given an XML template
   *
   * @param query The XML document used as the template for the search
   * @return The serviceEntry vector containing (if any) the matches
   *  that correspond to the query
   */
  public Vector querySET(Document query, String SETName) {
    //    System.out.println("Entering querySET(Doc, SETName): "+SETName);
    Hashtable rootHash = null;
    Treap rootTreap = null;
    Element root = query.getRoot();
    HashVector results = null;
    int i = root.numElements();
    //NOTE: replace & with whatever character is not allowed in tag names
    if (i > 1 || (i == 1 && root.getChild(0).getType() == Element.ELEMENT)) {
      //      System.out.println("1");
      rootHash = (Hashtable) rootTagList.get(SETName+"&"+SETserver.HASHSTORE);
      if (rootHash == null) return null;
      results = queryTag(rootHash, root);
      //      System.out.println("1");
    } else if (i == 1) {
      //    System.out.println("2");
      rootTreap = (Treap) rootTagList.get(SETName+"&"+SETserver.TREAPSTORE);
      if (rootTreap == null) return null;
      results = queryTag(rootTreap, root);
      //    System.out.println("2");
    } else {
      //    System.out.println("3");
      System.err.println("ERROR: impossible return from numElements()!");
      //    System.out.println("3");
      return null;
    }
    return totalIntersect(results);
  }

  /**
   * Method to do a group intersection of the total hashtables returned
   * by each tag search
   * TESTED!!
   */
  private Vector totalIntersect(HashVector hv) {
    //    System.out.println("Entering totalIntersect(hv)");
    if (!hv.check()) {
      System.err.println("ERROR: consistency check in hashVector failed!");
      return null;
    }
    Vector results = new Vector();
    if (hv.noresult) return results; // short circuit
    Hashtable counter = new Hashtable();
    Hashtable stored = new Hashtable();

    Enumeration hashtables = hv.hashtables.elements();
    Enumeration keys;
    String key;
    Hashtable hPtr;
    Integer count;
    while (hashtables.hasMoreElements()) {
      hPtr = (Hashtable) hashtables.nextElement();
      keys = hPtr.keys();
      while (keys.hasMoreElements()) {
	key = (String) keys.nextElement();
	stored.put(key, hPtr.get(key));
	count = (Integer) counter.get(key);
	if (count == null) count = new Integer(1);
	else {
	  hPtr.remove(key);
	  count = new Integer(count.intValue() +1);
	}
	counter.put(key, count);
      }
    }
    keys = counter.keys();
    int i ;
    while (keys.hasMoreElements()) {
      key = (String) keys.nextElement();
      i = ((Integer) counter.get(key)).intValue();
      if (i == hv.size) {
	results.addElement(stored.get(key));
      }
    }
    return results;
  }

  /**
   * Class to encapsulate the text and children tags
   * Right now it's not used, because the parser functionality is not there.
   * Tried to implement this and do a substring query on the current 
   * getText(), but that doesn't work, since substring does not allow for
   * missing words, etc etc.
   */
  private class textUnion {
    public String text;
    public Hashtable children;
    public textUnion(String t, Hashtable h) {
      text = t;
      children = h;
    }
  }

  /**
   * Class to keep track of multiple responses in hashtables.
   * Use this to return list of hashtables that we get from
   * range queries, and also to combine recursive results into larger sets
   */
  private class HashVector {
    public int size;
    public Vector hashtables;
    boolean noresult;
    public HashVector(Hashtable h) {
      if (h == null) {
	noresult = true;
	size = 0;
      } else {
        noresult = false;
	size = 1;
	hashtables = new Vector();
	hashtables.addElement(h);
      }
    }
    public HashVector() {
      noresult = false;
      size = 0;
      hashtables = new Vector();
    }
    public HashVector(boolean shortcircuit) {
      noresult = true;
      size = 0;
    }
    //function to check consistency inside the hashVector
    public boolean check() {
      if (noresult == false) return (size == hashtables.size());
      return true;
    }
    public void addHashtable(Hashtable h) {
      size++;
      hashtables.addElement(h);
    }
    /**
     * Adds the contents of hashVector hv to this hashVector
     */
    public void merge(HashVector hv) {
      if (hv.noresult == true || noresult == true) {
	noresult = true;
	return;
      }
      Enumeration e = hv.hashtables.elements();
      while (e.hasMoreElements()) {
	this.addHashtable((Hashtable) e.nextElement());
      }
    }
  }

  /**
   * Overloaded method that does a queryTag on an end node.  The data 
   * structure in question is a Treap.
   * @param t The treap that we're using to match up against the element
   * @param e The query node that should match the treap of values
   * @return A HashVector object that contains the hashtables
   * NOTE!!! ADD RANGE QUERY FUNCTIONALITY!!!!
   */
  private HashVector queryTag(Treap t, Element e) {
    //    System.out.println("Entering queryTag(Treap, Element): "+
    //		       e.getTagName().toString()+" "+e.getText());
    HashVector hv = null;
    Hashtable h = null;
    //This should be an end node, hence no children
    //    xset.assert(e.numElements() == 1);
    
    TreapOper to = null;;
    String att = null;
    //operator values includes whether range is inclusive
    //and the type of the value used as the key
    if ((att = (String) e.getAttribute("LESS_THAN")) != null) {
      System.out.println("U: DOING RANGE LOOKUP");
      to = new TreapOper();
      to.lessThan = att;
      to.lessEquals = false;
      e.removeAttribute("LESS_THAN");
    }
    if ((att = (String) e.getAttribute("MORE_THAN")) != null) {
      System.out.println("V: DOING RANGE LOOKUP");
      if (to == null) to = new TreapOper();
      to.moreThan = att;
      to.moreEquals = false;
      e.removeAttribute("MORE_THAN");
    }
    if ((att = (String) e.getAttribute("LESS_EQUAL")) != null) {
      System.out.println("W: DOING RANGE LOOKUP");
      if (to == null) to = new TreapOper();
      to.lessThan = att;
      to.lessEquals = true;
      e.removeAttribute("LESS_EQUAL");
    }
    if ((att = (String) e.getAttribute("MORE_EQUAL")) != null) {
      System.out.println("Y: DOING RANGE LOOKUP");
      if (to == null) to = new TreapOper();
      to.moreThan = att;
      to.moreEquals = true;
      e.removeAttribute("MORE_EQUAL");
    }
    if (to != null) {  // Range query exists
      if (to.lessThan == null && to.moreThan == null) {
	System.err.println("WARNING: range query has no bounds");
	h = (Hashtable) t.lookup(e.getText());
	return new HashVector(checkAttributes(h, e));
      }
      att = (String) e.getAttribute("KEY_TYPE");
      if (att.equalsIgnoreCase("STRING"))
	to.keyType = TreapOper.STRINGKEY;
      else if (att.equalsIgnoreCase("INTEGER"))
	to.keyType = TreapOper.INTKEY;
      else if (att.equalsIgnoreCase("FLOAT"))
	to.keyType = TreapOper.FLOATKEY;
      else System.err.println("ERROR: Invalid rangeQuery KeyType!");
      e.removeAttribute("KEY_TYPE");
      Vector tempV = t.rangeLookup(to); // Do range query
      System.out.println("Z: FINISHED RANGE LOOKUP");
      Enumeration enu = tempV.elements();
      Vector addHash = new Vector();
      while (enu.hasMoreElements()) {
	h = (Hashtable) enu.nextElement();
	addHash.addElement(h);
	// check attributes and then add to HashVector
      }
      // if no results after checkAttributes, then short circuit
      if ((h = checkAttributes(combineHash(addHash), e)) == null) {
	hv = new HashVector(false);
      } else {
        hv = new HashVector(h);
      }
    } else {   // no range queries, simple lookup
      h = (Hashtable) t.lookup(e.getText());
      // if no results after checkAttributes, then short circuit
      if ((h = checkAttributes(h, e)) == null) {
	hv = new HashVector(false);
      } else {
	hv = new HashVector(h);
      }
    }
    return hv;
  }

  private Hashtable combineHash(Vector v) {
    Enumeration e = v.elements(), keys, hashes;
    Hashtable total = new Hashtable(), h;
    String key;
    while (e.hasMoreElements()) {
      h = (Hashtable) e.nextElement();
      keys = h.keys();
      while (keys.hasMoreElements()) {
	key = (String) keys.nextElement();
	total.put(key, h.get(key));
      }
    }
    return total;
  }

  /**
   * Method to do the recursive XMLTag treap traversal
   * This only takes care of the recursive case.
   * NOTE: This isn't great, since it does non-end recursion
   *
   * @param t The treap that we're using to matching up against the element
   * @param e The element that should be matched up against the treap
   * @return A HashVector with the hashtable of serviceEntries that 
   *         match on the query
   */
  private HashVector queryTag(Hashtable topHash, Element e) {
    //    System.out.println("Entering queryTag(Hash, Element): "+
    //		       e.getTagName().toString()+" "+e.getText());
    Hashtable hPtr = null;
    Treap tPtr = null;
    HashVector hv = new HashVector();
    Element elem = null;
    Element currentE = null;
    String tagName;

    //    for (int i=0; (elem=e.getChildren().getChild(i))!=null; i++) {
    //      System.out.println("child #"+i+": "+elem.getText());
    //    }

    ElementCollection ec = e.getChildren();
    int j = ec.getLength();

    for (int i = 0; i < j; i++) {
      if (hv.noresult) return hv; // if noresult then short circuit
      currentE = ec.getChild(i);
      tagName = currentE.getTagName().toString();
      if (currentE.numElements() == 1 &&
	currentE.getChild(0).getType() != Element.ELEMENT) {
	tPtr = (Treap) topHash.get(tagName+"&"+SETserver.TREAPSTORE);
	if (tPtr != null) hv.merge(queryTag(tPtr, currentE));
	else {
	    // No documents found, so no match possible,
	    // short circuit the whole hashvector to give quick null result
	    hv.merge(new HashVector(false));
	}
      } else {
	hPtr = (Hashtable) topHash.get(tagName+"&"+SETserver.HASHSTORE);
	hv.merge(queryTag(hPtr, currentE));
      }
    }
    return hv;

    /* Here's a PROBLEM!!! currently no solution as to how to handle
     * attribute tags for matching attributes for nodes with children
     * elements.
     * TEMP. SOLN.: checkAttributes in queryTag(TREAPS, ELEMENTS)
     */
  }

  /**
   * Helper method used to do hashtable intersection
   * It takes entries from two hashtables, then puts their intersection
   * into a new hashtable.
   * @param a First vector
   * @param b Second vector
   * @return Hashtable a, now containing the intersection of old-a and b
  private Hashtable hashIntersect(Hashtable a, Hashtable b) { 
    Document d = null; 
    Enumeration e = a.keys(); 
    String key = null; 
    while (e.hasMoreElements()) { 
      key = (String) e.nextElement(); 
       check to see what keys of a exist in b
       for simplicity sake, assuming that message digest provides
       uniqueness in the value of the identifier key
       if key not in b, then remove from a
      if (!b.containsKey(key)) {
	a.remove(key);
      }
    }
    return a;
  }
   */

  /**
   * Method to check that all XML document entries in this Hashtable match
   * the attribute(s) that we are looking for.
   * @param v Vector that contains the documentEntries
   * @param e Element that we're matching attributes up against
   * @return a hashtable that contains the serviceEntrys which match
   */
  private Hashtable checkAttributes(Hashtable v, Element e) {
    //  System.out.println("Entering chckAttr(Hash, Element): "+e.getText());
    Hashtable ansV = new Hashtable();
    if (v == null) return (Hashtable) null;
    Enumeration docKeys = v.keys();
    DocumentEntry de = null;
    Document d = null;
    Attribute a = null;

    // put the Element attributes key value pairs into a hashtable
    Hashtable eleAttrs = new Hashtable();
    Enumeration ele = e.getAttributes();
    while (ele.hasMoreElements()) {
      a = (Attribute) ele.nextElement();
      System.out.println("attribute "+a.getName().toString()+" v: "+a.getValue());
      eleAttrs.put(a.getName().toString(), a.getValue());
    }

    Hashtable temphash = new Hashtable();
    Enumeration attrs = null;
    boolean matched;
    String currKey = null;
    while (docKeys.hasMoreElements()) {
      matched = true;
      currKey = (String) docKeys.nextElement();
      de = (DocumentEntry) v.get(currKey);
      attrs = de.attributes.elements();
      // put all of the attributes from this document entry into hashtable
      while (attrs.hasMoreElements()) {
	a = (Attribute) attrs.nextElement();
	temphash.put(a.getName().toString(), a.getValue());
      }
      Enumeration keys = eleAttrs.keys();
      String k = null;
      while (keys.hasMoreElements()) {
	k = (String) keys.nextElement();
	if (!((String)eleAttrs.get(k)).equalsIgnoreCase((String)temphash.get(k))) {
	  matched = false;
	  break;
	}
      }
      if (matched) {
	ansV.put(currKey, de.se);
      }
    }
    if (ansV.isEmpty()) return (Hashtable) null;
    return ansV;
  }

  /**
   * Method called by the SDS server to update an existing 
   * XML document.  It hashes the document to get the ID and
   * retrieves the serviceEntry, then updates the heartbeat field.
   * @param renewDoc XML document to be renewed
   */
  public void SETrenew(Document d) {
    String serviceID = null;
    ServiceEntry se = null;
    try {
      serviceID = SETutils.hasher(d, "SHA-1");
    } catch (java.security.NoSuchAlgorithmException nsae) {
      nsae.printStackTrace();
    }
    se = (ServiceEntry) globalHashSID.get(serviceID);
    se.updateTime(System.currentTimeMillis());
  }
    
  /** 
   * Method called by the SDS server or independent application to
   * replace an existing document with a new document.  It deletes
   * the old document first, the registers the new document.
   * @param olddoc The XML document to be replaced in the SET
   * @param newdoc The XML document to replace olddoc
   */
  public void SETreplace(Document d) {
    SETderegister(d);
    SETregister(d, "default");
  }

  /**
   * Method called by application to add new document to XSet.
   * This version is to provide a wrapper for parsing the XML string
   *
   * @param xmltext The XML document in string format
   * @param SETName The name of the XSet server being accessed
   */
  public void SETregister(String xmltext, String SETName) {
    Document d = new Document();
    ByteArrayInputStream bais = new ByteArrayInputStream(xmltext.getBytes());
    try {
	d.load(bais);
    } catch (Exception e) { e.printStackTrace(); return; }
    SETregister(d, SETName);
  }

  /**
   * Method called by the SDS server or independent application to
   * add a new document to the XSet search engine. Mainly a wrapper for
   * generating the serviceID
   *
   * @param d The XML document to be added to the search engine
   * @param SETName The name of the XSet server that is being accessed
   */
  public void SETregister(Document d, String SETName) {
    //    System.out.println("Entering SETregister(Doc, SETName): "+SETName);
    String serviceID = null;
    try {
      serviceID = SETutils.hasher(d, "SHA-1");
    } catch (java.security.NoSuchAlgorithmException nsae) {
      nsae.printStackTrace();
    }
    SETregister(d, SETName, serviceID);
  }

  /**
   * Method called by application to add new document to XSet.
   * This version is to provide a wrapper for parsing the XML string
   *
   * @param xmltext The XML document in string format
   * @param SETName The name of the XSet server being accessed
   */
  public void SETregister(String xmltext, String SETName, String serviceID) {
    Document d = new Document();
    ByteArrayInputStream bais = new ByteArrayInputStream(xmltext.getBytes());
    try {
	d.load(bais);
    } catch (Exception e) { e.printStackTrace(); return; }
    SETregister(d, SETName, serviceID);
  }

  /**
   * Method called by the SDS server or independent application to
   * add a new document to the XSet search engine.
   *
   * @param d The XML document to be added to the search engine
   * @param SETName The name of the XSet server that is being accessed
   * @param serviceID A previously generated ID for the document
   */
  public void SETregister(Document d, String SETName, String serviceID) {
    //    System.out.println("Entering SETregister(Doc, SETName): "+SETName);
    Treap rootTreap = null;
    Hashtable rootHash = null;
    Element root = d.getRoot();
    Name tagName = root.getTagName();
    int tType;

    ServiceEntry se = new ServiceEntry(serviceID, d, 
				       System.currentTimeMillis());
    globalHashSID.put(serviceID, se);
    int i = root.numElements();
    if ((i > 1) || 
	(i == 1 && root.getChild(0).getType() == Element.ELEMENT)) {
      // HAS CHILDREN
      rootHash = (Hashtable) rootTagList.get(SETName+"&"+SETserver.HASHSTORE);
      //      System.out.println("Found root Hashtable");
      if (rootHash == null) {
	rootHash = new Hashtable();
	rootTagList.put(SETName+"&"+SETserver.HASHSTORE, rootHash);
      }
      addToHashSeq(rootHash, se, root);
    } else {
      // NO CHILDREN
      rootTreap = (Treap) rootTagList.get(SETName+"&"+SETserver.TREAPSTORE);
      System.out.println("Found root Treap");
      if (rootTreap == null) {
	rootTreap = new Treap(SETName);
	rootTagList.put(SETName+"&"+SETserver.TREAPSTORE, rootTreap);
      }
      addToTreap(rootTreap, se, root);
    }
    d = null;
  }

  /**
   * Method called to take a serviceEntry, an element, and a treap,
   * and insert the element value into the treap, registering the document
   *
   * @param t The treap we're using to store the document hashtables under
   * @param se The service entry that encapsulates document d
   * @param elem The element in document d that we're filing under
   */
  private void addToTreap(Treap t, ServiceEntry se, Element elem) {
    //    System.out.println("Entering addToTreap(Treap,SE,Elem): "+
    //		       elem.getTagName().toString()+ " " + elem.getText());
    Hashtable hPtr;
    String tagValue;

    //    printTreap(t, "BEFORE");
    tagValue = elem.getText();
    // look for the given value for this tap in the treap
    if ((hPtr = (Hashtable) t.lookup(tagValue)) == null) {
      // no entry found, make new hashtable and insert into the Treap
      hPtr = new Hashtable();
      //      System.out.println("adding new hashtable to treap for tag: "+tagValue);
      hPtr.put(se.getid(), new DocumentEntry(se, elem.getAttributes()));
      try {
	t.insert(tagValue, hPtr);
      } catch (DuplicateException de) {
	de.printStackTrace();
      }
      // add new hashtable to global list of hashtables (for cleaner)
      //	globalHashList.addElement(hPtr);
      // add a backpointer of the hashtable to the serviceEntry
      se.addHashtable(hPtr);
    } else {
      //	hPtr.remove(se.getid());  // should be redundant
      hPtr.put(se.getid(), new DocumentEntry(se, elem.getAttributes()));
      // add a backpointer of the hashtable to the serviceEntry
      se.addHashtable(hPtr);
      //      printTreap(t, "AFTER");
    }
  }

  /**
   * Method called to take a document, its root element, and a Hashtable,
   * and traverse down the list of Hashtables, inserting the correct
   * tags down the tree of Hashtables.
   * @param h The current Hashtable we're looking at
   * @param se The service entry that encapsulates document d
   * @param root The tag in document d that we're filing under in t
   */
  private void addToHashSeq(Hashtable currHash, ServiceEntry se, Element elem){
    //    System.out.println("Entering addToHashSeq(Hash,SE,Elem): "+
    //		       elem.getTagName().toString()+ " " + elem.getText());
    Element ePtr = null;
    ElementCollection ecPtr = null;
    String tagValue = null;

    Treap tPtr = null;
    Hashtable hPtr = null;

    ecPtr = elem.getChildren();
    // take the list of children, and add them one by one
    int j = ecPtr.getLength();
    //    System.out.println("children #: "+j);
    for(int i=0; i<j; i++) {
      ePtr = ecPtr.getChild(i);
      //      System.out.println("child #: "+i+" "+ePtr.getText());
      //      System.out.println("  numElements: "+ePtr.numElements());
      int z = ePtr.numElements();
      if ((z > 1) || // child has children, => hashtable
	  (z == 1 && ePtr.getChild(0).getType() == Element.ELEMENT)) { 
	hPtr = (Hashtable) currHash.get(ePtr.getTagName().toString()+"&"+
					SETserver.HASHSTORE);
	if (hPtr == null) { // NOT FOUND
	  hPtr = new Hashtable();
	  currHash.put(ePtr.getTagName().toString()+"&"+SETserver.HASHSTORE,
		       hPtr);
	}
	addToHashSeq(hPtr, se, ePtr);
      } else {      // child has no children, => treap
	tPtr = (Treap) currHash.get(ePtr.getTagName().toString()+"&"+
				    SETserver.TREAPSTORE);
	if (tPtr == null) {
	  tPtr = new Treap(ePtr.getTagName().toString());
	  currHash.put(ePtr.getTagName().toString()+"&"+SETserver.TREAPSTORE,
		       tPtr);
	}
	addToTreap(tPtr, se, ePtr);
      }
    }
  }
	  //// STOPPED HERE, CONTINUE ON QUESTION of
	  //// SHOULD I ENCODE TREAPTYPE as part of the KEY
	  //// in TREAP insertion????

    /*
      for each top level child tag C of "root" in the current query, 
      execute add Query C to each sub-treap that matches C.  if C is
      not in the tag treap, then create a new treap, add value of C to
      it, and add a reference to that XML doc to that node.  If C has
      no more children then the treap that C is associated with had better
      contain only values, NOT other treap pointers.

      As a result, each treap must have a few more things associated with
      it.  It needs to contain: 1. what each node's key's type is,
      whether it be tagName (or String), so that nodes point to other
      treaps, or else the keys are the actual values of the tag, and
      the nodes point to XML vectors. 2.  if the treaps are of the latter
      type, we must know what the types of the keys are, and there needs
      to be some comparison operator defined and associated with the
      type.  ie if type ==integer, then you're ok.  If type == String,
      we need the same associated string comparator operators.
      */

  /**
   * Method called by the SDS server or independent application to
   * remove a document from the SET search engine indices.
   *
   * @param d The XML document to be removed from the search engine
   * @return true or false, whether the deregistration was successful
   */
  public void SETderegister(Document d) {
    System.out.println("Entering SETdereg(Document d)");
    String serviceID = null;
    try {
      serviceID = SETutils.hasher(d, "SHA-1");
    } catch (java.security.NoSuchAlgorithmException nsae) {
      nsae.printStackTrace();
      return;
    }
    ServiceEntry se = (ServiceEntry) globalHashSID.get(serviceID);
    delete(se);
  }
  
  /**
   * Method called by SETderegister and the cleaner to remove
   * a serviceEntry from the SET server.
   * It looks up the hashtables pointing to it, and removes the
   * references from them one by one.  It also removes the reference
   * from the master ServiceID to ServiceEntry hashtable.
   * @param se The serviceEntry to be deleted.
   */
  public void delete(ServiceEntry se) {
    System.out.println("Entering delete(ServiceEntry): "+se.toString());
    // get list of associated hashtables from the serviceEntry
    globalHashSID.remove(se.getid());
    Enumeration ePtr = se.getHashList().elements();
    Hashtable hPtr = null;
    while (ePtr.hasMoreElements()) {
      // pull up the hashtable, and delete the se reference from it
      hPtr = (Hashtable) ePtr.nextElement();
      hPtr.remove(se.getid());
      if (hPtr.isEmpty()) {
	//NEED TO ADD CODE TO DELETE THE EMPTY HASHTABLE!!!!!!!!
      }
    }
  }

  /**
   * TESTER Method for checkAttributes
   */
  public void testCheckAttributes() {
    Document d = new Document();
    try {
      d.load(new FileInputStream("e.xml"));
    } catch (Exception e) {}
    Element root = d.getRoot();
    ElementCollection ec = root.getChildren();
    Element e = ec.getChild(0);
    
    // now need to crate the DocumentENtry with attributes
    ServiceEntry se = new ServiceEntry("ID", new Document(), 0);
    Element f = ec.getChild(1);
    DocumentEntry de = new DocumentEntry(se, f.getAttributes());
    Hashtable h = new Hashtable();
    h.put("a", de);
    Hashtable ans = checkAttributes(h, e);
    if (ans.isEmpty()) {
      System.out.println("No matches");
    } else {
      System.out.println("Matched!");
    }
  }

  /**
   * TESTER Method for totalIntersect
   */
  public void testIntersect () {
    HashVector hv = new HashVector();
    int i;
    Hashtable h = new Hashtable();
    for(i=1; i<10; i++) {
      h.put(String.valueOf(i), new Integer(i));
    }
    hv.addHashtable(h);
    h = new Hashtable();
    for (i=8; i<25; i++) {
      h.put(String.valueOf(i), new Integer(i));
    }
    hv.addHashtable(h);
    h = new Hashtable();
    for (i=3; i<15; i++) {
      h.put(String.valueOf(i), new Integer(i));
    }
    hv.addHashtable(h);
    Vector x = totalIntersect(hv);
    Enumeration e = x.elements();
    while (e.hasMoreElements()) {
      System.out.println(e.nextElement().toString());
    }
  }
  /**
   * Debugging method used to print out the registration treap
   * It uses the Treap.print()
   */
  public void printSET(String setname) {
    System.out.println("Entering printSET(SETName): "+setname);
    Hashtable roota = null;
    Treap rootb = null;
    Enumeration keys = rootTagList.keys();
    String k = null;
    while (keys.hasMoreElements()) {
      k = (String) keys.nextElement();
      System.out.println("Key in hashtable: "+k);
    }
    roota = (Hashtable) rootTagList.get(setname+"&"+SETserver.HASHSTORE);
    rootb = (Treap) rootTagList.get(setname+"&"+SETserver.TREAPSTORE);
    System.out.println("ENTERING PRINTER");
    if (roota != null) {
      printHash(roota, "");
    }
    if (rootb != null) {
      printTreap(rootb, "");
    }
  }

  /**
   * Method used to recursively travel down the SET internal structure
   * and print out information for debugging purposes
   * @param h The hashtable to traverse/print
   * @param offset The space offset to printout the tag with
   */
  private void printHash(Hashtable h, String offset) {
    Enumeration keys = h.keys();
    String key;
    Object obj;
    while(keys.hasMoreElements()) {
      key = (String) keys.nextElement();
      System.out.println("Tag: "+offset+key);
      obj = h.get(key);
      if (obj instanceof Hashtable) {
	printHash((Hashtable) obj, offset+"   ");
      } else if (obj != null) {
	printTreap((Treap) obj, offset+"   ");
      }
    }
  }
  /**
   * Method used to recursively travel down the SET internal structure
   * and print out information for debugging purposes
   * @param t The Treap to traverse/print
   */
  private void printTreap(Treap t, String offset) {
    System.out.println("Printing Treap");
    Enumeration e = t.keys();
    String key;
    Hashtable h;
    while(e.hasMoreElements()) {
      key = (String) e.nextElement();
      h = (Hashtable) t.lookup(key);
      printH(h, offset, key);
    }
  }
  private void printH (Hashtable h, String offset, String key) {
    Enumeration hashe = h.keys();
    String hashkey;
    DocumentEntry de;
    while (hashe.hasMoreElements()) {
      hashkey = (String) hashe.nextElement();
      de = (DocumentEntry) h.get(hashkey);
      if (de == null || de.se == null) 
	System.out.println("EMPTY DOCUMENTENTRY");
      System.out.println(offset+"tag: "+key+" k: "+hashkey+ 
			 "\n"+offset+" id: "+de.se.getid()+"\n v: "+
			 de.se.getXML());
    }
  }

}

